using System;
using System.IO;
using NUnit.Framework;

namespace Smdn.Formats.Thbgm {
  [TestFixture]
  public class TrackInfoTests {
    [Test]
    public void TestCalculateRepeatedTimes()
    {
      // intro: 0.25sec, repeat: 1.00sec
      var trackInfo = new TrackInfo(null, "test", StreamFormat.ThXX, 0x10, 44100L, 176400L);

      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.00)));
      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.25)));

      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(1.25 - 0.125)));
      Assert.AreEqual(1, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(1.25)));
      Assert.AreEqual(1, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(1.25 + 0.125)));

      Assert.AreEqual(1, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(2.25 - 0.125)));
      Assert.AreEqual(2, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(2.25)));
      Assert.AreEqual(2, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(2.25 + 0.125)));
    }

    [Test]
    public void TestRepeatLengthZeroCalculateRepeatedTimes()
    {
      // intro: 0.25sec, repeat: 0.00sec
      var trackInfo = new TrackInfo(null, "test", StreamFormat.ThXX, 0x10, 44100L, 0L);

      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.00)));
      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.25)));
      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.50)));
      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(0.75)));
      Assert.AreEqual(0, trackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(1.00)));
    }

    [Test]
    public void TestCalculateRepeatPositionInBytes()
    {
      // intro: 0.25sec, repeat: 1.00sec
      var trackInfo = new TrackInfo(null, "test", StreamFormat.ThXX, 0x10, 44100L, 176400L);

      Assert.AreEqual(44100L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(0, false));
      Assert.AreEqual(44100L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(0, true));

      Assert.AreEqual(44100L + 176400L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(1, false));
      Assert.AreEqual(44100L + 176400L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(1, true));

      Assert.AreEqual(44100L + 2 * 176400L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(2, false));
      Assert.AreEqual(44100L + 2 * 176400L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(2, true));
    }

    [Test]
    public void TestRepeatLengthZeroCalculateRepeatPositionInBytes()
    {
      // intro: 0.25sec, repeat: 0.00sec
      var trackInfo = new TrackInfo(null, "test", StreamFormat.ThXX, 0x10, 44100L, 0L);

      Assert.AreEqual(44100L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(0, false));
      Assert.AreEqual(44100L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(0, true));

      Assert.AreEqual(44100L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(1, false));
      Assert.AreEqual(44100L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(1, true));

      Assert.AreEqual(44100L + 0L,
                      trackInfo.CalculateRepeatPositionInBytes(2, false));
      Assert.AreEqual(44100L + 0x10,
                      trackInfo.CalculateRepeatPositionInBytes(2, true));
    }

    [Test]
    public void TestThplayFormatProductInfoSamplingRateFieldSpecified()
    {
      var testfilename = "titles_thplay.txt";

      try {
        if (File.Exists(testfilename))
          File.Delete(testfilename);

        File.WriteAllText(testfilename, @"@骨川アリス幻樂団\東方臑茶魔体験版\thbgm_tr.dat,東方臑茶魔　～ The Acquaintance Of My Daddy. 体験版
#=SamplingRate,22050,99999999,99999999,99999999,スネ夫が激しい弾幕を展開する時に流れている曲1
99999999,99999999,99999999,スネ夫が激しい弾幕を展開する時に流れている曲2
");

        var product = ProductInfo.LoadFrom(testfilename);

        Assert.AreEqual(2, product.Tracks.Count);

        var track = product.Tracks[0];

        Assert.AreEqual(22050, track.StreamFormat.SamplesPerSecond);
        Assert.AreEqual(16, track.StreamFormat.BitsPerSample);
        Assert.AreEqual(2, track.StreamFormat.Channels);
        Assert.AreEqual("スネ夫が激しい弾幕を展開する時に流れている曲1", track.Title);

        track = product.Tracks[1];

        Assert.AreEqual(44100, track.StreamFormat.SamplesPerSecond);
        Assert.AreEqual(16, track.StreamFormat.BitsPerSample);
        Assert.AreEqual(2, track.StreamFormat.Channels);
        Assert.AreEqual("スネ夫が激しい弾幕を展開する時に流れている曲2", track.Title);
      }
      finally {
        if (File.Exists(testfilename))
          File.Delete(testfilename);
      }
    }

    [Test]
    public void TestThplayFormatProductInfoStreamFileFieldSpecified()
    {
      var testfilename = "titles_thplay.txt";

      try {
        if (File.Exists(testfilename))
          File.Delete(testfilename);

        File.WriteAllText(testfilename, @"@骨川アリス幻樂団\東方臑茶魔体験版\thbgm_tr.dat,東方臑茶魔　～ The Acquaintance Of My Daddy. 体験版
%track1.wav,99999999,99999999,99999999,スネ夫が激しい弾幕を展開する時に流れている曲1
99999999,99999999,99999999,スネ夫が激しい弾幕を展開する時に流れている曲2
");

        var product = ProductInfo.LoadFrom(testfilename);

        Assert.AreEqual(2, product.Tracks.Count);

        var track = product.Tracks[0];

        Assert.AreEqual(44100, track.StreamFormat.SamplesPerSecond);
        Assert.AreEqual(16, track.StreamFormat.BitsPerSample);
        Assert.AreEqual(2, track.StreamFormat.Channels);
        Assert.AreEqual("スネ夫が激しい弾幕を展開する時に流れている曲1", track.Title);

        try {
          using (var s = track.GetStream("/tmp/thplay/stream/", 2)) {
            Assert.Fail("expected IOException not thrown");
          }
        }
        catch (Exception ex) {
          var fnfex = ex as FileNotFoundException;
          var dnfex = ex as DirectoryNotFoundException;

          if (fnfex != null)
            Assert.AreEqual("/tmp/thplay/stream/track1.wav", fnfex.FileName);
          else if (dnfex != null)
            StringAssert.Contains("/tmp/thplay/stream/track1.wav", dnfex.Message);
          else
            throw;
        }

        track = product.Tracks[1];

        Assert.AreEqual(44100, track.StreamFormat.SamplesPerSecond);
        Assert.AreEqual(16, track.StreamFormat.BitsPerSample);
        Assert.AreEqual(2, track.StreamFormat.Channels);
        Assert.AreEqual("スネ夫が激しい弾幕を展開する時に流れている曲2", track.Title);

        try {
          using (var s = track.GetStream("/tmp/thplay/stream/stream.dat", 2)) {
            Assert.Fail("expected IOException not thrown");
          }
        }
        catch (Exception ex) {
          var fnfex = ex as FileNotFoundException;
          var dnfex = ex as DirectoryNotFoundException;

          if (fnfex != null)
            Assert.AreEqual("/tmp/thplay/stream/stream.dat", fnfex.FileName);
          else if (dnfex != null)
            StringAssert.Contains("/tmp/thplay/stream/stream.dat", dnfex.Message);
          else
            throw;
        }
      }
      finally {
        if (File.Exists(testfilename))
          File.Delete(testfilename);
      }
    }
  }
}
