// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.ComponentModel;

using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class Effect : ICloneable, INotifyPropertyChanged {
    public static TimeSpan TimeSpanUIMaxValue = TimeSpan.FromHours(6.0);

    public static string TimeSpanToStringConverter(TimeSpan val)
    {
      return string.Format("{0}:{1:D2}:{2:D2}.{3:D3}", (int)val.TotalHours, val.Minutes, val.Seconds, val.Milliseconds);
    }

    public static TimeSpan StringToTimeSpanConverter(string val)
    {
      var values = val.Split(new[] {System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.TimeSeparator}, StringSplitOptions.None);

      if (values.Length <= 0)
        throw new FormatException();

      var secs = double.Parse(values[values.Length - 1]);

      switch (values.Length) {
        case 1: { // sss.fff
          return TimeSpan.FromSeconds(secs);
        }

        case 2: { // mm:ss.fff
          var mins = (double)int.Parse(values[0]);
          return TimeSpan.FromSeconds(mins * 60.0 + secs);
        }

        case 3: { // hh:mm:ss.ff
          var hours = (double)int.Parse(values[0]);
          var mins  = (double)int.Parse(values[1]);
          return TimeSpan.FromSeconds(hours * 3600.0 + mins * 60.0 + secs);
        }

        default:
          throw new FormatException();
      }
    }

    public event PropertyChangedEventHandler PropertyChanged;

    public TimeSpan FadingStart {
      get { return fadingStart; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("FadingStart", "must be zero or positive value");

        fadingStart = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingStart"));
      }
    }

    public TimeSpan FadingLength {
      get { return fadingLength; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("FadingLength", "must be zero or positive value");

        fadingLength = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingLength"));
      }
    }

    public TimeSpan TrailingSilenceLength {
      get { return trailingSilenceLength; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("TrailingSilenceLength", "must be zero or positive value");

        trailingSilenceLength = value;

        OnPropertyChanged(new PropertyChangedEventArgs("TrailingSilenceLength"));
      }
    }

    public FadingCurveType FadingCurveType {
      get { return fadingCurveType; }
      set
      {
        fadingCurveType = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingCurveType"));
      }
    }

    public TimeSpan TotalLength {
      get { return fadingStart + fadingLength + trailingSilenceLength; }
    }

    public static Effect CreateDefault(TrackInfo track)
    {
      if (track == null)
        throw new ArgumentNullException("track");

      var effect = new Effect();

      // 2 repeats, 20sec fading, 0sec silence
      effect.SetLength(track, 2, TimeSpan.FromSeconds(20.0), TimeSpan.FromSeconds(-20.0), TimeSpan.Zero);

      // linear curve fading
      effect.FadingCurveType = FadingCurveType.Linear;

      return effect;
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public virtual Effect Clone()
    {
      var cloned = (Effect)MemberwiseClone();

      cloned.PropertyChanged = null;

      return cloned;
    }

    public void SetLength(TimeSpan fadingStart, TimeSpan fadingLength, TimeSpan trailingSilenceLength)
    {
      FadingStart = fadingStart;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    public void SetTotalLength(TimeSpan totalLength, TimeSpan fadingLength, TimeSpan trailingSilenceLength)
    {
      var start = totalLength - fadingLength - trailingSilenceLength;

      if (start < TimeSpan.Zero)
        start = TimeSpan.Zero;

      FadingStart = start;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    public void SetLength(TrackInfo track, int timesToRepeat, TimeSpan fadingLength, TimeSpan fadingOffset, TimeSpan trailingSilenceLength)
    {
      var start = track.CalculateRepeatPosition(timesToRepeat) + fadingOffset;

      if (start < TimeSpan.Zero)
        start = TimeSpan.Zero;

      FadingStart = start;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }

    private TimeSpan fadingStart;
    private TimeSpan fadingLength;
    private TimeSpan trailingSilenceLength;
    private FadingCurveType fadingCurveType;
  }
}
