// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;

using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class EffectAppliedBgmStream : BgmStream {
    public Effect Effect {
      get { return effect; }
      set
      {
        if (value == null)
          throw new ArgumentNullException("value");
        effect = value;
      }
    }

    public override long Length {
      get
      {
        if (constructed)
          return AlignOffset(base.Format.BytesPerSecond * (effect.FadingStart + effect.FadingLength + effect.TrailingSilenceLength).TotalSeconds);
        else
          // when called by constructor
          return base.Length;
      }
    }

    public override long Position {
      get { return base.Position - bufferRemain; }
      set { base.Position = value; DiscardBuffer(); }
    }

    public EffectAppliedBgmStream(string path, TrackInfo trackInfo, int timesToRepeat, Effect effect)
      : this(path, trackInfo, timesToRepeat, effect, trackInfo.StreamFormat.BytesPerSecond)
    {
    }

    public EffectAppliedBgmStreamPart Part {
      get { return part; }
    }

    public EffectAppliedBgmStream(string path, TrackInfo trackInfo, int timesToRepeat, Effect effect, int bufferSize)
      : base(path, trackInfo, Repeat.Infinite) // ignore value of timesToRepeat param
    {
      if (bufferSize % trackInfo.StreamFormat.BlockAlign != 0)
        throw new ArgumentException("must be n times of block align", "bufferSize");

      constructed = true;

      this.buffer = new byte[bufferSize];
      this.bufferOffset = 0;
      this.bufferRemain = 0;
      this.Effect = effect;
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      var currentPosition = Position;

      DiscardBuffer();

      switch (origin) {
        case SeekOrigin.Begin:
          return base.Seek(AlignOffset(offset), SeekOrigin.Begin);
        case SeekOrigin.Current:
          return base.Seek(AlignOffset(currentPosition + offset), SeekOrigin.Begin);
        case SeekOrigin.End:
          throw new NotSupportedException();
        default:
          throw new System.ComponentModel.InvalidEnumArgumentException("origin", (int)origin, typeof(SeekOrigin));
      }
    }

    public override int Read(byte[] array, int offset, int count)
    {
      if (array == null)
        throw new ArgumentNullException("array");
      if (count < 0)
        throw new ArgumentOutOfRangeException("count", "must be zero or positive number");
      if (offset < 0)
        throw new ArgumentOutOfRangeException("offset", "must be zero or positive number");
      if (array.Length < count + offset)
        throw new ArgumentException("array.Length < count + offset");

      var read = 0;

      for (;;) {
        if (bufferRemain == 0) {
          if (eofReached)
            return read;

          FillBuffer(ref eofReached);
        }

        if (count <= bufferRemain) {
          Buffer.BlockCopy(buffer, bufferOffset, array, offset, count);

          bufferOffset += count;
          bufferRemain -= count;

          return read + count;
        }
        else {
          Buffer.BlockCopy(buffer, bufferOffset, array, offset, bufferRemain);

          offset += bufferRemain;
          count  -= bufferRemain;
          read   += bufferRemain;

          bufferRemain = 0;
        }
      }
    }

    private void DiscardBuffer()
    {
      bufferOffset = buffer.Length;
      bufferRemain = 0;
      eofReached = false;
    }

    private long TimeSpanToPosition(TimeSpan timeSpan)
    {
      return AlignOffset(timeSpan.TotalSeconds * base.Format.BytesPerSecond);
    }

    private long AlignOffset(double offset)
    {
      return (long)(offset / Format.BlockAlign) * Format.BlockAlign;
    }

    private long AlignOffset(long offset)
    {
      return (offset / Format.BlockAlign) * Format.BlockAlign;
    }

    private void FillBuffer(ref bool eof)
    {
      if (eof) {
        bufferOffset = buffer.Length;
        bufferRemain = 0;
        part = EffectAppliedBgmStreamPart.EndOfStream;
        return;
      }

      var bufferingOffset = base.Position;

      bufferOffset = 0;
      bufferRemain = base.Read(buffer, 0, buffer.Length);

      if (bufferRemain < buffer.Length)
        eof = true;

      var fadingStartOffset = TimeSpanToPosition(effect.FadingStart);

      if (bufferingOffset + bufferRemain < fadingStartOffset) {
        if (RepeatedTimes <= 0)
          part = EffectAppliedBgmStreamPart.Intro;
        else
          part = EffectAppliedBgmStreamPart.Loop;
        return;
      }

      // fading
      var fadingEndOffset = fadingStartOffset + TimeSpanToPosition(effect.FadingLength);

      switch (effect.FadingCurveType) {
        case FadingCurveType.Linear:
        case FadingCurveType.Exponential:
        case FadingCurveType.Logarithmic:
        case FadingCurveType.Trigonometrical: {
          var fadingPosition = bufferingOffset - fadingStartOffset;
          var fadingLength = (double)(fadingEndOffset - fadingStartOffset);

          if (base.Format.Channels == 2 && base.Format.BitsPerSample == 16) {
            unsafe {
              fixed (byte* ptr = buffer) {
                var wave = (byte*)ptr;

                for (var i = 0; i < bufferRemain; i += 4, fadingPosition += 4) {
                  if (fadingPosition < 0) {
                    wave += 4;
                    continue;
                  }
                  else if (fadingLength <= fadingPosition) {
                    break;
                  }

                  var factor = 1.0 - fadingPosition / fadingLength;

                  if (effect.FadingCurveType == FadingCurveType.Exponential)
                    // f(x) = (n^x - 1.0) / (n - 1.0) where n = 10
                    factor = (Math.Pow(10.0, factor) - 1.0) / 9.0;
                  else if (effect.FadingCurveType == FadingCurveType.Logarithmic)
                    // f(x) = log_n(x * (n - 1.0) + 1.0) where n = 10
                    factor = Math.Log10(factor * 9.0 + 1.0);
                  else if (effect.FadingCurveType == FadingCurveType.Trigonometrical)
                    // f(x) = (1.0 - cos(x * π)) / 2.0
                    factor = (1.0 - Math.Cos(factor * Math.PI)) * 0.5;
                  /*
                  else if (effect.FadingCurveType == FadingCurveType.Linear)
                    // f(x) = x
                    factor = factor
                  */

                  var left  = (ushort)((short)(wave[0] | (wave[1] << 8)) * factor);
                  var right = (ushort)((short)(wave[2] | (wave[3] << 8)) * factor);

                  *(wave++) = (byte)(left & 0x00ff);
                  *(wave++) = (byte)(left >> 8);
                  *(wave++) = (byte)(right & 0x00ff);
                  *(wave++) = (byte)(right >> 8);
                }
              }
            }
          }
          else {
            throw new NotImplementedException();
          }
          break;
        }

        default:
          // no effect
          break;
      } // switch

      if (bufferingOffset + bufferRemain < fadingEndOffset) {
        part = EffectAppliedBgmStreamPart.FadeOut;
        return;
      }

      // trailing silence
      var silenceOffset = fadingEndOffset - bufferingOffset;

      if (silenceOffset < 0) {
        Array.Clear(buffer, 0, buffer.Length);
      }
      else {
        for (var i = silenceOffset; i < bufferRemain; i++) {
          buffer[i] = 0;
        }
      }

      var trailingSilenceEndOffset = fadingEndOffset + TimeSpanToPosition(effect.TrailingSilenceLength);

      if (bufferingOffset + bufferRemain < trailingSilenceEndOffset) {
        part = EffectAppliedBgmStreamPart.TrailingSilence;
        return;
      }

      // eof
      bufferRemain = (int)(trailingSilenceEndOffset - bufferingOffset);

      if (bufferRemain < 0)
        bufferRemain = 0;

      part = EffectAppliedBgmStreamPart.EndOfStream;

      eof = true;
    }

    private bool constructed = false;
    private byte[] buffer;
    private int bufferOffset = 0;
    private int bufferRemain = 0;
    private bool eofReached;
    private Effect effect;
    private EffectAppliedBgmStreamPart part = EffectAppliedBgmStreamPart.Intro;
  }
}
