// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;
using Smdn.Media;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public partial class Player : UserControl {
    /*
     * class members
     */
    public const int MaxVolume = 100;
    public const int MinVolume = 0;

    static Player()
    {
      imageButtonPlay = new Bitmap(24, 24);
      imageButtonStop = new Bitmap(24, 24);
      imageButtonPause = new Bitmap(24, 24);

      var images = new[] {
        imageButtonPlay, imageButtonStop, imageButtonPause,
      };

      for (var i = 0; i < images.Length; i++) {
        var image = images[i];

        using (var g = Graphics.FromImage(image)) {
          g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
          g.Clear(Color.Transparent);

          switch (i) {
            case 0:
              var trianglePoints = new[] {
                new PointF(image.Width * 0.2f, image.Height * 0.2f),
                new PointF(image.Width * 0.2f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.5f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints);
              break;

            case 1:
              var rectangle = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, rectangle);
              break;

            case 2:
              var leftRectangle  = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.4f, image.Height * 0.8f);
              var rightRectangle = RectangleF.FromLTRB(image.Width * 0.6f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, leftRectangle);
              g.FillRectangle(SystemBrushes.ControlText, rightRectangle);
              break;
          }
        }
      }
    }

    private static Bitmap imageButtonPlay;
    private static Bitmap imageButtonStop;
    private static Bitmap imageButtonPause;

    /*
     * events
     */
    public event EventHandler PlayDone;
    public event EventHandler TrackChanged;

    /*
     * properties
     */
    public TimeSpan Position {
      get
      {
        RejectDisposed();
        if (player == null)
          return TimeSpan.MinValue;
        else
          return player.Position;
      }
      set
      {
        RejectDisposed();
        if (player != null)
          player.Position = value;
      }
    }

    public Track Track {
      get
      {
        RejectDisposed();
        return track;
      }
      set
      {
        RejectDisposed();
        OnTrackChanged(value);
      }
    }

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set
      {
        if (value == bgmSourcePath)
          return;

        bgmSourcePath = value;

        Stop();

        buttonPlay.Enabled = File.Exists(bgmSourcePath);

        waveFormView.BgmSourcePath = BgmSourcePath;
      }
    }

    public int Volume {
      get { return trackBarVolume.Value; }
      set { trackBarVolume.Value = value; }
    }

    public bool Repeat {
      get { return checkBoxRepeat.Checked; }
      set
      {
        RejectDisposed();

        checkBoxRepeat.Checked = value;

        if (player != null)
          player.Repeat = value;
      }
    }

    /*
     * construction/destruction
     */
    public Player()
    {
      InitializeComponent();

      player = PcmStreamPlayer.Create();
      player.PlayDone += delegate {
        DisableStopButton();
        OnPlayDone(EventArgs.Empty);
      };

      waveFormView.SeekBarMoved += delegate(object sender, SeekBarMovedEventArgs e) {
        if (player == null || !player.Playing)
          e.Cancel = true;
        else
          player.Position = e.SeekBarPosition;
      };

      trackBarVolume.Minimum = MinVolume;
      trackBarVolume.Maximum = MaxVolume;
      trackBarVolume.SmallChange = (MaxVolume - MinVolume) / 100;
      trackBarVolume.LargeChange = (MaxVolume - MinVolume) / 10;
      trackBarVolume.Value = (MaxVolume - MinVolume) * 3 / 4;
      trackBarVolume.ValueChanged += delegate {
        SetPlayerVolume();
      };

      buttonPlay.Click += delegate {
        if (player == null || !player.Playing)
          Play();
        else if (player.Pausing)
          Resume();
        else
          Pause();
      };

      buttonPlay.Image = imageButtonPlay;
      buttonPlay.ImageAlign = ContentAlignment.MiddleCenter;

      buttonStop.Click += delegate { Stop(); };
      buttonStop.Image = imageButtonStop;
      buttonStop.ImageAlign = ContentAlignment.MiddleCenter;
      buttonStop.Enabled = false;

      checkBoxRepeat.CheckedChanged += delegate {
        if (player != null)
          player.Repeat = checkBoxRepeat.Checked;
      };

      SetPlayerVolume();
    }

    protected override void OnLoad(EventArgs e)
    {
      timer = new System.Threading.Timer(delegate(object state) {
        SetSeekBarPosition();
      }, null, 100, 100);

      base.OnLoad(e);
    }

    private void DisposeInternal(bool disposing)
    {
      if (disposing) {
        if (!disposed) {
          Stop();

          if (timer != null) {
            timer.Dispose();
            timer = null;
          }

          if (player != null) {
            player.Close();
            player = null;
          }

          if (stream != null) {
            stream.Close();
            stream = null;
          }

          disposed = true;
        }
      }
    }

    /*
     * public methods
     */
    public void Play()
    {
      RejectDisposed();

      if (track == null)
        return;

      buttonPlay.Image = imageButtonPause;
      buttonStop.Enabled = true;

      try {
        stream = track.GetStream(BgmSourcePath);
      }
      catch (FileNotFoundException) {
        throw;
      }
      catch (ArgumentException) {
        buttonPlay.Image = imageButtonPlay;
        buttonStop.Enabled = false;
        return;
      }

      SetPlayerVolume();

      player.PlayAsync(stream, stream.Format.GetWaveFormatEx(), checkBoxRepeat.Checked);
    }

    public void Stop()
    {
      RejectDisposed();

      if (player == null)
        return;

      DisableStopButton();

      player.Stop();
    }

    public void Pause()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPlay;

      player.Pause();
    }

    public void Resume()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPause;

      player.Resume();
    }

    /*
     * private methods
     */
    private void SetPlayerVolume()
    {
      if (player == null)
        return;

      var scale = (trackBarVolume.Value - trackBarVolume.Minimum) / (double)(trackBarVolume.Maximum - trackBarVolume.Minimum);

      player.Volume = (int)(100 * scale);
    }

    private void DisableStopButton()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)DisableStopButton);
        return;
      }

      buttonPlay.Image = imageButtonPlay;
      buttonStop.Enabled = false;
    }

    private void SetSeekBarPosition()
    {
      if (player == null)
        return;

      if (InvokeRequired) {
        Invoke((MethodInvoker)SetSeekBarPosition);
      }
      else {
        var position = player.Position;

        labelPosition.Text = string.Format("{0}:{1:D2}.{2:D3}", (int)position.TotalMinutes, position.Seconds, position.Milliseconds);

        waveFormView.SeekBarPosition = position;
      }
    }

    private void OnTrackChanged(Track newTrack)
    {
      if (track == newTrack)
        return;

      Stop();

      if (stream != null) {
        stream.Close();
        stream = null;
      }

      track = newTrack;

      waveFormView.Track = track;

      effectEditor.Effect = (track == null) ? null : track.Effect;

      OnTrackChanged(EventArgs.Empty);
    }

    protected virtual void OnTrackChanged(EventArgs e)
    {
      var ev = this.TrackChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPlayDone(EventArgs e)
    {
      var ev = this.PlayDone;

      if (ev != null)
        ev(this, e);
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private PcmStreamPlayer player;
    private Track track = null;
    private string bgmSourcePath = null;
    private EffectAppliedBgmStream stream = null;
    private System.Threading.Timer timer = null;
  }
}
