﻿// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;
using System.Windows.Forms;

using Smdn.IO;
using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public partial class ProductSelector : UserControl {
    public event EventHandler SelectedProductChanged;
    public event EventHandler BgmSourcePathChanged;

    public IEnumerable<ProductInfo> Products {
      get { return products; }
      set
      {
        if (this.products == value)
          return;

        this.products = value;

        productInfoDropDownList.Items.Clear();

        if (products != null) {
          productInfoDropDownList.BeginUpdate();
          foreach (var product in products) {
            productInfoDropDownList.Items.Add(product);
          }
          productInfoDropDownList.EndUpdate();

          if (0 < productInfoDropDownList.Items.Count)
            productInfoDropDownList.SelectedIndex = productInfoDropDownList.Items.Count - 1;
        }
      }
    }

    public ProductInfo SelectedProduct {
      get
      {
        if (productInfoDropDownList.SelectedIndex < 0)
          return null;
        else
          return productInfoDropDownList.SelectedItem;
      }
      set
      {
        foreach (var item in productInfoDropDownList.Items) {
          if (item == value) {
            productInfoDropDownList.SelectedItem = item;
            return;
          }
        }
      }
    }

    public string BgmSourcePath {
      get { return textBoxBgmSourcePath.Text; }
      set { SetBgmSourcePath(value, false); }
    }

    private bool AutoDetected {
      get { return autoDetected; }
      set
      {
        autoDetected = value;

        productInfoDropDownList.Enabled = !autoDetected;
      }
    }

    public ProductSelector()
    {
      InitializeComponent();

      this.Products = new ProductInfo[] {};

      this.productInfoDropDownList.SelectedIndexChanged += delegate {
        OnSelectedProductChanged(EventArgs.Empty);
      };

      this.buttonChangeBgmSourcePath.Click += buttonChangeBgmSourcePath_Click;

      this.textBoxBgmSourcePath.TextChanged += delegate {
        OnBgmSourcePathChanged(EventArgs.Empty);
      };
    }

    public void FocusBgmSourcePath()
    {
      textBoxBgmSourcePath.Focus();
      textBoxBgmSourcePath.SelectAll();
    }

    protected override void OnDragEnter(DragEventArgs drgevent)
    {
      base.OnDragEnter(drgevent);

      if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
        drgevent.Effect = DragDropEffects.All;
      else
        drgevent.Effect = DragDropEffects.None;
    }

    protected override void OnDragDrop(DragEventArgs drgevent)
    {
      try {
        var files = (string[])drgevent.Data.GetData(DataFormats.FileDrop, false);

        if (0 < files.Length)
          SetBgmSourcePath(files[0], true);
      }
      catch {
        // ignore exceptions
      }
      finally {
        base.OnDragDrop(drgevent);
      }
    }

    public void ShowChangeBmgSourcePathDialog()
    {
      using (var dialog = new OpenFileDialog()) {
        if (SelectedProduct == ProductInfo.Th06)
          dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
            new[] {Texts.GetText("BgmSourceFile"), "th06_*.wav", "*.dat"},
            new[] {Texts.GetText("AllFiles"), "*.*"},
          });
        else
          dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
            new[] {Texts.GetText("BgmSourceFile"), "*.dat"},
            new[] {Texts.GetText("AllFiles"), "*.*"},
          });

        var path = textBoxBgmSourcePath.Text;

        if (string.IsNullOrEmpty(path)) {
          if (SelectedProduct == null)
            path = Environment.CurrentDirectory;
          else
            path = SelectedProduct.DefaultBgmSourcePath;
        }

        dialog.FilterIndex = 0;
        dialog.FileName = path;
        dialog.InitialDirectory = Path.GetDirectoryName(path);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) == DialogResult.OK)
          SetBgmSourcePath(dialog.FileName, true);
      }
    }

    private void buttonChangeBgmSourcePath_Click(object sender, EventArgs e)
    {
      ShowChangeBmgSourcePathDialog();
    }

    public void SetBgmSourcePath(string newBgmSourcePath, bool autoDetect)
    {
      if (autoDetect) {
        var product = ProductInfo.FindMatchedProduct(newBgmSourcePath, products);

        this.AutoDetected = (product != null);

        textBoxBgmSourcePath.Text = newBgmSourcePath;

        this.SelectedProduct = product;
      }
      else {
        this.AutoDetected = false;

        textBoxBgmSourcePath.Text = newBgmSourcePath;
      }
   }

    protected virtual void OnSelectedProductChanged(EventArgs e)
    {
      var ev = this.SelectedProductChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnBgmSourcePathChanged(EventArgs e)
    {
      var ev = this.BgmSourcePathChanged;

      if (ev != null)
        ev(this, e);
    }

    private bool autoDetected = false;
    private IEnumerable<ProductInfo> products;
  }
}
