﻿// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class TrackListView : ListView {
    public event CancelEventHandler ProductChanging;
    public event EventHandler ProductChanged;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        var e = new CancelEventArgs();

        OnProductChanging(e);

        if (e.Cancel)
          return;

        product = value;

        OnProductChanged();
      }
    }

    public Track SelectedTrack {
      get
      {
        if (SelectedItems.Count <= 0)
          return null;
        else
          return (SelectedItems[0] as TrackListViewItem).Track;
      }
    }

    public IEnumerable<Track> Tracks {
      get { return tracks; }
    }

    public TrackListView()
      : base()
    {
      CheckBoxes = true;
      View = View.Details;
      HeaderStyle = ColumnHeaderStyle.Nonclickable;
      MultiSelect = false;
      FullRowSelect = true;
      HideSelection = false;

      Columns.Add(Texts.GetText("TrackNumber"), 45);
      Columns.Add(Texts.GetText("Title"), 290);
#if DEBUG
      Columns.Add("IsDirty", 20);
#endif
      Columns.Add(Texts.GetText("TotalLength"), 100);
      Columns.Add(Texts.GetText("RepeatedTimes"), 70);
      Columns.Add(Texts.GetText("FadingStart"), 100);
      Columns.Add(Texts.GetText("FadingLength"), 140);
      Columns.Add(Texts.GetText("FadingCurveType"), 120);
      Columns.Add(Texts.GetText("TrailingSilenceLength"), 100);
    }

    public void SetProductAndTracks(ProductInfo product, IEnumerable<Track> tracks)
    {
      itemsUpdating = true;

      BeginUpdate();

      Items.Clear();

      this.tracks = new List<Track>();

      foreach (var track in tracks) {
        var item = new TrackListViewItem(track);

        item.Checked = track.Checked;

        Items.Add(item);

        this.tracks.Add(track);
      }

      EndUpdate();

      itemsUpdating = false;

      if (this.product != product) {
        this.product = product;

        // raise event
        OnProductChanged(EventArgs.Empty);
      }
    }

    protected override void OnItemChecked(ItemCheckedEventArgs e)
    {
      var item = (e.Item as TrackListViewItem);

      item.Track.Checked = item.Checked;

      if (!itemsUpdating)
        // edit by user operation
        item.Track.SetDirty();

      base.OnItemChecked(e);
    }

    private void OnProductChanged()
    {
      itemsUpdating = true;

      BeginUpdate();

      Items.Clear();

      tracks = new List<Track>();

      if (product != null) {
        foreach (var trackInfo in product.Tracks) {
          var track = new Track(trackInfo);

          tracks.Add(track);

          var item = new TrackListViewItem(track);

          item.Checked = track.Checked;

          Items.Add(item);
        }
      }

      EndUpdate();

      itemsUpdating = false;

      OnProductChanged(EventArgs.Empty);
    }

    protected virtual void OnProductChanging(CancelEventArgs e)
    {
      var ev = this.ProductChanging;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnProductChanged(EventArgs e)
    {
      var ev = this.ProductChanged;

      if (ev != null)
        ev(this, e);
    }

    private ProductInfo product = null;
    private List<Track> tracks = new List<Track>();
    private bool itemsUpdating = false;
  }
}
