using System;
using System.Collections.Generic;
using System.ComponentModel;

using Smdn.Formats.Ini;
using Smdn.Formats.Thbgm;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public class ExtractionConfig : ICloneable, INotifyPropertyChanging, INotifyPropertyChanged {
    public static string FileExtension = ".ini";

    public event PropertyChangingEventHandler PropertyChanging;
    public event PropertyChangedEventHandler PropertyChanged;

    private ProductInfo product;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        if (OnPropertyChanging("Product")) {
          product = value;

          OnPropertyChanged("Product");
        }
      }
    }

    private string bgmSourcePath;

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set
      {
        if (bgmSourcePath == value)
          return;

        if (OnPropertyChanging("BgmSourcePath")) {
          bgmSourcePath = value;

          OnPropertyChanged("BgmSourcePath");
        }
      }
    }

    private string outputDirectory;

    public string OutputDirectory {
      get { return outputDirectory; }
      set
      {
        if (outputDirectory == value)
          return;

        if (OnPropertyChanging("OutputDirectory")) {
          outputDirectory = value;

          OnPropertyChanged("OutputDirectory");
        }
      }
    }

    private bool prependTrackNumber = true;

    public bool PrependTrackNumber {
      get { return prependTrackNumber; }
      set
      {
        if (prependTrackNumber == value)
          return;

        if (OnPropertyChanging("PrependTrackNumber")) {
          prependTrackNumber = value;

          OnPropertyChanged("PrependTrackNumber");
        }
      }
    }

    private bool useWaveDashInsteadOfFullWidthTilde = false;

    public bool UseWaveDashInsteadOfFullWidthTilde {
      get { return useWaveDashInsteadOfFullWidthTilde; }
      set
      {
        if (useWaveDashInsteadOfFullWidthTilde == value)
          return;

        if (OnPropertyChanging("UseWaveDashInsteadOfFullWidthTilde")) {
          useWaveDashInsteadOfFullWidthTilde = value;

          OnPropertyChanged("UseWaveDashInsteadOfFullWidthTilde");
        }
      }
    }

    private bool useTwoSpacesInsteadOfIdeographicSpace = false;

    public bool UseTwoSpacesInsteadOfIdeographicSpace {
      get { return useTwoSpacesInsteadOfIdeographicSpace; }
      set
      {
        if (useTwoSpacesInsteadOfIdeographicSpace == value)
          return;

        if (OnPropertyChanging("UseTwoSpacesInsteadOfIdeographicSpace")) {
          useTwoSpacesInsteadOfIdeographicSpace = value;

          OnPropertyChanged("UseTwoSpacesInsteadOfIdeographicSpace");
        }
      }
    }

    private EncoderConfig encoderConfig = EncoderConfig.Empty;

    public EncoderConfig EncoderConfig {
      get { return encoderConfig; }
      set
      {
        if (encoderConfig == value)
          return;

        if (OnPropertyChanging("EncoderConfig")) {
          encoderConfig = value;

          OnPropertyChanged("EncoderConfig");
        }
      }
    }

    private bool saveSourceAfterEncode = false;

    public bool SaveSourceAfterEncode {
      get { return saveSourceAfterEncode; }
      set
      {
        if (saveSourceAfterEncode == value)
          return;

        if (OnPropertyChanging("SaveSourceAfterEncode")) {
          saveSourceAfterEncode = value;

          OnPropertyChanged("SaveSourceAfterEncode");
        }
      }
    }

    private IEnumerable<Track> tracks = new Track[] {};

    public IEnumerable<Track> Tracks {
      get { return tracks; }
      set
      {
        if (tracks == value)
          return;

        if (OnPropertyChanging("Tracks")) {
          tracks = value;

          OnPropertyChanged("Tracks");
        }
      }
    }

    private Dictionary<Track, string> extractedFilePaths = new Dictionary<Track, string>();

    public Dictionary<Track, string> ExtractedFilePaths {
      get { return extractedFilePaths; }
    }

    private Dictionary<Track, string> encodedFilePaths = new Dictionary<Track, string>();

    public Dictionary<Track, string> EncodedFilePaths {
      get { return encodedFilePaths; }
    }

    public TitleUnificationOptions TitleUnificationOptions {
      get
      {
        var options = TitleUnificationOptions.NoUnify;

        if (useTwoSpacesInsteadOfIdeographicSpace)
          options |= TitleUnificationOptions.DoubleSpace;
        else
          options |= TitleUnificationOptions.SingleIdeographicSpace;

        if (useWaveDashInsteadOfFullWidthTilde)
          options |= TitleUnificationOptions.WaveDash;
        else
          options |= TitleUnificationOptions.FullWidthTilde;

        return options;
      }
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public virtual ExtractionConfig Clone()
    {
      var cloned = (ExtractionConfig)MemberwiseClone();

      cloned.extractedFilePaths = new Dictionary<Track, string>(this.extractedFilePaths);
      cloned.encodedFilePaths = new Dictionary<Track, string>(this.encodedFilePaths);

      cloned.PropertyChanged = null;
      cloned.PropertyChanging = null;

      return cloned;
    }

    public void Save(string file)
    {
      var document = new IniDocument();

      document.DefaultSection["Version"] = System.Windows.Forms.Application.ProductVersion;

      var extractionConfigSection = document["ExtractionConfig"];

      extractionConfigSection["Product"]                = product.Title;
      extractionConfigSection["BgmSourcePath"]          = bgmSourcePath;
      extractionConfigSection["OutputDirectory"]        = outputDirectory;
      extractionConfigSection["PrependTrackNumber"]     = prependTrackNumber.ToString();
      extractionConfigSection["UseWaveDashInsteadOfFullWidthTilde"]     = useWaveDashInsteadOfFullWidthTilde.ToString();
      extractionConfigSection["UseTwoSpacesInsteadOfIdeographicSpace"]  = useTwoSpacesInsteadOfIdeographicSpace.ToString();
      extractionConfigSection["SaveSourceAfterEncode"]  = saveSourceAfterEncode.ToString();
      extractionConfigSection["EncoderConfig"]          = EncoderConfig.IsNullOrEmpty(encoderConfig) ? string.Empty : encoderConfig.Name;

      foreach (var track in tracks) {
        var trackSection = document[string.Format("Track{0}", track.TrackInfo.TrackNumber)];

        //trackSection["TrackNumber"] = track.TrackInfo.TrackNumber.ToString();
        trackSection["Checked"]     = track.Checked.ToString();

        trackSection["FadingStart"]           = track.Effect.FadingStart.ToString();
        trackSection["FadingLength"]          = track.Effect.FadingLength.ToString();
        trackSection["FadingCurveType"]       = track.Effect.FadingCurveType.ToString();
        trackSection["TrailingSilenceLength"] = track.Effect.TrailingSilenceLength.ToString();
      }

      document.Save(file);
    }

    public static ExtractionConfig LoadFrom(string file)
    {
      try {
        return LoadFromCore(file);
      }
      catch (NotSupportedException) {
        throw;
      }
      catch (Exception ex) {
        throw new System.IO.InvalidDataException("invalid extraction config", ex);
      }
    }

    private static ExtractionConfig LoadFromCore(string file)
    {
      var document = IniDocument.Load(file);
      var config = new ExtractionConfig();

      //var version = document.DefaultSection["Version"];

      var extractionConfigSection = document["ExtractionConfig"];

      config.product = null;

      foreach (var product in ThbgmExtractor.Products) {
        if (product.Title == extractionConfigSection["Product"])
          config.product = product;
      }

      if (config.product == null)
        throw new NotSupportedException("unsupported product");

      config.bgmSourcePath   = extractionConfigSection["BgmSourcePath"];
      config.outputDirectory = extractionConfigSection["OutputDirectory"];

      config.prependTrackNumber                     = extractionConfigSection.GetValue<bool?>("PrependTrackNumber", NullableConvert.ToBooleanIgnoreException) ?? config.prependTrackNumber;
      config.useWaveDashInsteadOfFullWidthTilde     = extractionConfigSection.GetValue<bool?>("UseWaveDashInsteadOfFullWidthTilde", NullableConvert.ToBooleanIgnoreException) ?? config.useWaveDashInsteadOfFullWidthTilde;
      config.useTwoSpacesInsteadOfIdeographicSpace  = extractionConfigSection.GetValue<bool?>("UseTwoSpacesInsteadOfIdeographicSpace", NullableConvert.ToBooleanIgnoreException) ?? config.useTwoSpacesInsteadOfIdeographicSpace;
      config.saveSourceAfterEncode                  = extractionConfigSection.GetValue<bool?>("PrependTrackNumber", NullableConvert.ToBooleanIgnoreException) ?? config.saveSourceAfterEncode;

      config.encoderConfig = EncoderConfig.Empty;

      foreach (var encoderConfig in ThbgmExtractor.EncoderConfigs) {
        if (encoderConfig.Name == extractionConfigSection["EncoderConfig"])
          config.encoderConfig = encoderConfig;
      }

      var tracks = new List<Track>();

      foreach (var trackInfo in config.product.Tracks) {
        var trackSection = document.Find(string.Format("Track{0}", trackInfo.TrackNumber));

        Track track;

        if (trackSection == null) {
          track = new Track(trackInfo);
        }
        else {
          var effect = Effect.CreateDefault(trackInfo);

          TimeSpan fadingStart;
          TimeSpan fadingLength;
          TimeSpan trailingSilenceLength;

          if (TimeSpan.TryParse(trackSection["FadingStart"], out fadingStart))
            effect.FadingStart = (fadingStart < TimeSpan.Zero) ? TimeSpan.Zero : fadingStart;
          if (TimeSpan.TryParse(trackSection["FadingLength"], out fadingLength))
            effect.FadingLength = (fadingLength < TimeSpan.Zero) ? TimeSpan.Zero : fadingLength;
          if (TimeSpan.TryParse(trackSection["TrailingSilenceLength"], out trailingSilenceLength))
            effect.TrailingSilenceLength = (trailingSilenceLength < TimeSpan.Zero) ? TimeSpan.Zero : trailingSilenceLength;

          effect.FadingCurveType = trackSection.GetValue<FadingCurveType?>("FadingCurveType", NullableConvert.ToEnumIgnoreException<FadingCurveType>) ?? effect.FadingCurveType;

          track = new Track(trackInfo, effect, bool.Parse(trackSection["Checked"]));
        }

        tracks.Add(track);
      }

      config.tracks = tracks;

      return config;
    }

    private bool OnPropertyChanging(string propertyName)
    {
      var e = new CancellablePropertyChangingEventArgs(propertyName);

      OnPropertyChanging(e);

      return !e.Cancel;
    }

    protected virtual void OnPropertyChanging(PropertyChangingEventArgs e)
    {
      var ev = this.PropertyChanging;

      if (ev != null)
        ev(this, e);
    }

    private void OnPropertyChanged(string propertyName)
    {
      OnPropertyChanged(new PropertyChangedEventArgs(propertyName));
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }
  }
}
