using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Smdn.IO;
using Smdn.Formats.Thbgm;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public static class Extractor {
    private class EncodingContext {
      public Queue<Track> EncodingTrackQueue {
        get; private set;
      }

      public ExtractionContext ExtractionContext {
        get; private set;
      }

      public EncodingContext(ExtractionContext extractionContext, Queue<Track> encodingTrackQueue)
      {
        this.ExtractionContext = extractionContext;
        this.EncodingTrackQueue = encodingTrackQueue;
      }
    }

    public static ExtractionContext StartExtract(ExtractionConfig config)
    {
      var context = new ExtractionContext(config);
      var t = new Thread(new ParameterizedThreadStart(Extract));

      context.ExtractorThread = t;

      t.Start(context);

      return context;
    }

    public static bool WaitForExtractionFinished(ExtractionContext context, int millisecondsTimeout)
    {
      return context.ExtractorThread.Join(millisecondsTimeout);
    }

    public static void Abort(ExtractionContext context)
    {
      if (context == null)
        throw new ArgumentNullException("context");

      context.AbortEvent.Set();
    }

    private static void Extract(object state)
    {
      var context = state as ExtractionContext;

      Extract(context);

      if (context.IsAbortRequested)
        UpdateProgressStatus(context.Progress, ThbgmExtractor.GetText("Extractor_Aborted"));
      else
        UpdateProgressStatus(context.Progress, ThbgmExtractor.GetText("Extractor_Completed"));
    }

    private static void Extract(ExtractionContext context)
    {
      var progress = context.Progress;
      var doEncode = !EncoderConfig.IsNullOrEmpty(context.Config.EncoderConfig);

      if (context.IsAbortRequested)
        return;

      var encodingTrackQueue = new Queue<Track>();
      var encodingContext = new EncodingContext(context, encodingTrackQueue);
      var encodingThread = doEncode ? new Thread((ParameterizedThreadStart)Encode) : null;

      // start encoding thread
      if (doEncode) {
        encodingThread.IsBackground = true;
        encodingThread.Start(encodingContext);
      }

      // start extraction
      UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_StartExtraction"));

      context.Stage = ExtractionStage.Extracting;

      foreach (var track in context.Config.Tracks) {
        try {
          Extract(context, track);

          if (context.IsAbortRequested)
            return;

          lock ((encodingTrackQueue as System.Collections.ICollection).SyncRoot) {
            encodingTrackQueue.Enqueue(track);
          }
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);
        }

        progress.ExtractedTrackCount++;
      }

      lock ((encodingTrackQueue as System.Collections.ICollection).SyncRoot) {
        encodingTrackQueue.Enqueue(null /* as terminator */);
      }

      if (context.IsAbortRequested) {
        if (doEncode)
          encodingThread.Abort();
        return;
      }

      context.Stage = ExtractionStage.Encoding;

      // wait for encoding thread finished
      if (doEncode) {
        UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_WaitForEncodingFinished"));
        encodingThread.Join();
      }

      context.Stage = ExtractionStage.Finished;
    }

    private static void Extract(ExtractionContext context, Track track)
    {
      if (context.IsAbortRequested)
        return;

      var config = context.Config;
      var progress = context.Progress;
      var outputFilePath = config.ExtractedFilePaths[track];

      UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_StartExtractTrack", progress.ExtractedTrackCount + 1, progress.TotalTrackCount, track.TrackInfo.Title));

      if (File.Exists(outputFilePath))
        // overwrite is allowed
        File.Delete(outputFilePath);

      UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractingTrack", track.TrackInfo.Title, outputFilePath));

      using (var inputStream = track.GetStream(config.BgmSourcePath)) {
        var lastPart = EffectAppliedBgmStreamPart.EndOfStream;
        var lastRepeatedTimes = inputStream.RepeatedTimes;

        progress.TrackLength = inputStream.Length;
        progress.ExtractedLength = 0L;

        using (var outputStream = File.OpenWrite(outputFilePath)) {
          var header = WaveStream.CreateRiffWaveHeaderFrom(inputStream);

          outputStream.SetLength(header.Length + inputStream.Length);
          outputStream.Write(header, 0, header.Length);

          var data = new byte[inputStream.Format.BytesPerSecond];

          for (;;) {
            if (context.IsAbortRequested)
              return;

            if (inputStream.Part != lastPart || inputStream.RepeatedTimes != lastRepeatedTimes) {
              lastPart          = inputStream.Part;
              lastRepeatedTimes = inputStream.RepeatedTimes;

              switch (lastPart) {
                case EffectAppliedBgmStreamPart.Intro:            UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractingIntroPart")); break;
                case EffectAppliedBgmStreamPart.Loop:             UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractingLoopPart", lastRepeatedTimes)); break;
                case EffectAppliedBgmStreamPart.FadeOut:          UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractingFadeOutPart", lastRepeatedTimes)); break;
                case EffectAppliedBgmStreamPart.TrailingSilence:  UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractingTrailingSilence")); break;
                default: break;
              }
            }

            var read = inputStream.Read(data, 0, data.Length);

            if (read <= 0)
              break;

            outputStream.Write(data, 0, read);

            progress.ExtractedLength += read;
          }

          outputStream.Flush();
        }
      }

      UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_ExtractTrackFinished", track.TrackInfo.Title));
    }

    private static void Encode(object state)
    {
      var context = state as EncodingContext;
      var extractionContext = context.ExtractionContext;
      var progress = extractionContext.Progress;

      try {
        progress.EncodedTrackCount = progress.TotalTrackCount;
        progress.CompletedTrackCount = progress.TotalTrackCount;

        var encoderConfig = extractionContext.Config.EncoderConfig;

        if (EncoderConfig.IsNullOrEmpty(encoderConfig))
          return;

        progress.EncodedTrackCount = 0;
        progress.CompletedTrackCount = 0;

        for (;;) {
          if (extractionContext.IsAbortRequested)
            return;

          Thread.Sleep(500);

          Track encodingTrack = null;

          // キューを監視し、キューに入ってきたエントリから順にエンコードする
          lock ((context.EncodingTrackQueue as System.Collections.ICollection).SyncRoot) {
            if (0 < context.EncodingTrackQueue.Count)
              encodingTrack = context.EncodingTrackQueue.Dequeue();
            else
              continue;
          }

          if (encodingTrack == null) // terminator
            return;

          try {
            UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_StartEncoding", progress.EncodedTrackCount + 1, progress.TotalTrackCount, encodingTrack.TrackInfo.Title));

            Encode(extractionContext, encodingTrack, encoderConfig);

            progress.EncodedTrackCount++;

            progress.CompletedTrackCount++;
          }
          catch (Exception ex) {
            Console.Error.WriteLine(ex);
          }
        }
      }
      catch (ThreadAbortException) {
        Console.Error.WriteLine(ThbgmExtractor.GetText("Extractor_EncodingAborted"));
      }
    }

    private static void Encode(ExtractionContext context, Track track, EncoderConfig encoderConfig)
    {
      var progress = context.Progress;

      var inputFilePath   = context.Config.ExtractedFilePaths[track];
      var outputFilePath  = context.Config.EncodedFilePaths[track];
      var encoderInputFilePath  = inputFilePath;
      var encoderOutputFilePath = outputFilePath;

      if (File.Exists(outputFilePath))
        // overwrite is allowed
        File.Delete(outputFilePath);

      var escapeShellSpecialChars = PathUtils.ContainsShellEscapeChar(inputFilePath) || PathUtils.ContainsShellEscapeChar(outputFilePath);

      if (escapeShellSpecialChars) {
        encoderInputFilePath = PathUtils.RenameUnique(inputFilePath);
        encoderOutputFilePath = Path.ChangeExtension(encoderInputFilePath, encoderConfig.OutputExtension);

        if (File.Exists(encoderOutputFilePath))
          File.Delete(encoderOutputFilePath);
      }

      Process encoderProcess = null;

      try {
        var psi = Smdn.Shell.CreateProcessStartInfo(encoderConfig.ExecutableFile, encoderConfig.Arguments);

        //psi.RedirectStandardInput = true;
        psi.RedirectStandardOutput = true;
        psi.RedirectStandardError = true;
        psi.UseShellExecute = false; // redirecting stdout and stderr
        psi.WorkingDirectory = Path.GetDirectoryName(Application.ExecutablePath);
#if DEBUG
        psi.CreateNoWindow = false;
#else
        psi.CreateNoWindow = true;
#endif

        psi.EnvironmentVariables.Add("THBGM_TITLE", track.TrackInfo.GetUnifiedTitle(context.Config.TitleUnificationOptions));
        psi.EnvironmentVariables.Add("THBGM_TRACK", track.TrackInfo.TrackNumber.ToString());
        psi.EnvironmentVariables.Add("THBGM_ALBUM", track.TrackInfo.Product.GetUnifiedTitle(context.Config.TitleUnificationOptions));
        psi.EnvironmentVariables.Add("THBGM_CREATOR", track.TrackInfo.Product.Creator);
        psi.EnvironmentVariables.Add("THBGM_RELEASE_YEAR", track.TrackInfo.Product.ReleaseDate.Year.ToString("D4"));
        psi.EnvironmentVariables.Add("THBGM_RELEASE_MONTH", track.TrackInfo.Product.ReleaseDate.Month.ToString("D2"));
        psi.EnvironmentVariables.Add("THBGM_RELEASE_DAY", track.TrackInfo.Product.ReleaseDate.Day.ToString("D2"));
        psi.EnvironmentVariables.Add("THBGM_INPUTFILEPATH", encoderInputFilePath);
        psi.EnvironmentVariables.Add("THBGM_OUTPUTFILEPATH", encoderOutputFilePath);

        using (encoderProcess = new Process()) {
          // start encoder process
          encoderProcess.StartInfo = psi;
          encoderProcess.OutputDataReceived += delegate(object sender, DataReceivedEventArgs e) {
            if (e.Data != null)
              Console.Out.WriteLine(e.Data);
          };
          encoderProcess.ErrorDataReceived += delegate(object sender, DataReceivedEventArgs e) {
            if (e.Data != null)
              Console.Error.WriteLine(e.Data);
          };

          encoderProcess.Start();
          encoderProcess.BeginOutputReadLine();
          encoderProcess.BeginErrorReadLine();

          Console.WriteLine("  encoder command line: {0} {1}", encoderProcess.StartInfo.FileName, encoderProcess.StartInfo.Arguments);
          Console.WriteLine("  THBGM_INPUTFILEPATH : {0}", encoderInputFilePath);
          Console.WriteLine("  THBGM_OUTPUTFILEPATH: {0}", encoderOutputFilePath);

          UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_EncodingTrack", track.TrackInfo.Title));

          while (!encoderProcess.HasExited) {
            if (context.IsAbortRequested) {
              UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_AbortedEncodingTrack", track.TrackInfo.Title));

              encoderProcess.Kill();
              encoderProcess = null;

              return;
            }

            encoderProcess.WaitForExit(2000);

            //encoderProcess.StandardInput.WriteLine("y");
          }
        }

        encoderProcess = null;

        UpdateProgressStatus(progress, ThbgmExtractor.GetText("Extractor_EncodeTrackFinished", track.TrackInfo.Title));

        if (escapeShellSpecialChars) {
          if (File.Exists(encoderOutputFilePath))
            File.Move(encoderOutputFilePath, outputFilePath);
          if (File.Exists(encoderInputFilePath))
            File.Move(encoderInputFilePath, inputFilePath);
        }

        if (File.Exists(outputFilePath) && !context.Config.SaveSourceAfterEncode)
          File.Delete(inputFilePath);
      }
      catch (System.ComponentModel.Win32Exception ex) {
        Console.Error.WriteLine(ex);
      }
      catch (ThreadAbortException) {
        if (encoderProcess != null && !encoderProcess.HasExited)
          encoderProcess.Kill();
      }
    }

    private static void UpdateProgressStatus(ExtractionProgress progress, string status)
    {
      progress.Status = status;

      Console.WriteLine("{0}: {1}", DateTime.Now.ToString("o"), progress.Status);
    }
  }
}
