using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class MainDialog : Form {
    private class EncoderConfigToolStripMenuItem : ToolStripMenuItem {
      public EncoderConfig EncoderConfig {
        get; private set;
      }

      public EncoderConfigToolStripMenuItem(EncoderConfig encoderConfig, EventHandler onClick)
        : base(encoderConfig.Name, null, onClick)
      {
        this.EncoderConfig = encoderConfig;
      }
    }

    private class FadingCuverTypeStripMenuItem : ToolStripMenuItem {
      public FadingCurveType FadingCurveType {
        get; private set;
      }

      public FadingCuverTypeStripMenuItem(FadingCurveType fadingCurveType, EventHandler onClick)
        : base(Texts.GetText(fadingCurveType), null, onClick)
      {
        this.FadingCurveType = fadingCurveType;
      }
    }

    private class TimesToRepeatStripMenuItem : ToolStripMenuItem {
      public int TimesToRepeat {
        get; private set;
      }

      public TimesToRepeatStripMenuItem(int timesToRepeat, EventHandler onClick)
        : base(ThbgmExtractor.GetText("TimesToRepeatStripMenuItem_FormatString", timesToRepeat), null, onClick)
      {
        this.TimesToRepeat = timesToRepeat;
      }
    }

    public MainDialog()
    {
      InitializeComponent();
      Initialize();

      if (Runtime.IsRunningOnMono) {
        // XXX
        this.ClientSize = new Size(784, 563);
        productSelector.Width = this.ClientSize.Width;
        splitContainerMain.Size = new Size(784, 486);
        splitContainerMain.SplitterDistance = 300;
        splitContainerExtraction.Size = new Size(784, 241);
        splitContainerExtraction.SplitterDistance = 490;
      }
    }

    private void Initialize()
    {
      textBoxLog.MaxLength = int.MaxValue;

      timerLog.Start();

      splitContainerExtraction.Panel2Collapsed = true;

      productSelector.SelectedProductChanged += delegate {
        edittingExtractionConfig.Product = productSelector.SelectedProduct;

        trackListView.Product = edittingExtractionConfig.Product;
      };

      productSelector.BgmSourcePathChanged += delegate {
        edittingExtractionConfig.BgmSourcePath = productSelector.BgmSourcePath;

        player.BgmSourcePath = edittingExtractionConfig.BgmSourcePath;
      };

      var sortedByReleaseDateProducts = new List<ProductInfo>(ThbgmExtractor.Products);

      sortedByReleaseDateProducts.Sort(delegate(ProductInfo x, ProductInfo y) {
        return -1/*negate, latest first*/ * DateTime.Compare(x.ReleaseDate, y.ReleaseDate);
      });

      productSelector.Products = sortedByReleaseDateProducts;

      var encoderConfigInsertIndex = toolStripMenuItemExtract_EncoderConfigs.DropDownItems.IndexOf(toolStripMenuItemExtract_EncoderConfigs_Separator);
      var encoderConfigOnClick = new EventHandler(delegate(object sender, EventArgs e) {
        edittingExtractionConfig.EncoderConfig = (sender as EncoderConfigToolStripMenuItem).EncoderConfig;
        isEdittingExtractionConfigDirty = true;

        SetSelectedEncoderConfigCheckedState();
      });

      foreach (var encoderConfig in ThbgmExtractor.EncoderConfigs) {
        toolStripMenuItemExtract_EncoderConfigs.DropDownItems.Insert(encoderConfigInsertIndex, new EncoderConfigToolStripMenuItem(encoderConfig, encoderConfigOnClick));

        encoderConfigInsertIndex++;
      }

      trackListView.ProductChanged += delegate {
        edittingExtractionConfig.Tracks = trackListView.Tracks;
        player.Track = null;
      };
      trackListView.SelectedIndexChanged += delegate {
        player.Track = trackListView.SelectedTrack;
      };

      trackListView.Product = null;
      trackListView.Product = productSelector.SelectedProduct;

      toolStripMenuItemExtract_PrependTrackNumber.Click += delegate {
        edittingExtractionConfig.PrependTrackNumber = toolStripMenuItemExtract_PrependTrackNumber.Checked;
        isEdittingExtractionConfigDirty = true;
      };

      toolStripMenuItemExtract_UseWaveDashInsteadOfFullWidthTilde.Click += delegate {
        edittingExtractionConfig.UseWaveDashInsteadOfFullWidthTilde = toolStripMenuItemExtract_UseWaveDashInsteadOfFullWidthTilde.Checked;
        isEdittingExtractionConfigDirty = true;
      };

      toolStripMenuItemExtract_UseTwoSpacesInsteadOfIdeographicSpace.Click += delegate {
        edittingExtractionConfig.UseTwoSpacesInsteadOfIdeographicSpace = toolStripMenuItemExtract_UseTwoSpacesInsteadOfIdeographicSpace.Checked;
        isEdittingExtractionConfigDirty = true;
      };

      toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Click += delegate {
        edittingExtractionConfig.SaveSourceAfterEncode = toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Checked;
        isEdittingExtractionConfigDirty = true;
      };

      foreach (var fadingCurveType in new[] {
        FadingCurveType.Linear,
        FadingCurveType.Exponential,
        FadingCurveType.Logarithmic,
        FadingCurveType.Trigonometrical,
        FadingCurveType.None,
      }) {
        toolStripMenuItemEdit_FadingCurveType.DropDownItems.Add(new FadingCuverTypeStripMenuItem(fadingCurveType, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack != null)
            trackListView.SelectedTrack.Effect.FadingCurveType = (sender as FadingCuverTypeStripMenuItem).FadingCurveType;
        }));
      }

      toolStripMenuItemEdit_FadingCurveType.DropDownOpening += delegate {
        var selectedFadingCurveType = (trackListView.SelectedTrack == null)
          ? (FadingCurveType?)null
          : (FadingCurveType?)trackListView.SelectedTrack.Effect.FadingCurveType;

        foreach (FadingCuverTypeStripMenuItem item in toolStripMenuItemEdit_FadingCurveType.DropDownItems) {
          item.Checked = (item.FadingCurveType == selectedFadingCurveType);
        }
      };

      for (var i = 0; i < 10; i++) {
        toolStripMenuItemEdit_TimesToRepeat.DropDownItems.Add(new TimesToRepeatStripMenuItem(i, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack == null)
            return;

          var track = trackListView.SelectedTrack;
          var fadingOffset = track.Effect.FadingStart - track.TrackInfo.CalculateRepeatPosition(track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength));

          track.Effect.SetLength(track.TrackInfo, (sender as TimesToRepeatStripMenuItem).TimesToRepeat, track.Effect.FadingLength, fadingOffset, track.Effect.TrailingSilenceLength);
        }));
      }

      toolStripMenuItemEdit_TimesToRepeat.DropDownOpening += delegate {
        var selectedTimesToRepeat = (trackListView.SelectedTrack == null)
          ? -1
          : trackListView.SelectedTrack.TrackInfo.CalculateRepeatedTimes(trackListView.SelectedTrack.Effect.TotalLength);

        foreach (TimesToRepeatStripMenuItem item in toolStripMenuItemEdit_TimesToRepeat.DropDownItems) {
          item.Checked = (item.TimesToRepeat == selectedTimesToRepeat);
        }
      };

      toolStripMenuItemEdit_SelectAll.Click += delegate {
        foreach (TrackListViewItem item in trackListView.Items) {
          item.Checked = true;
        }
      };

      toolStripMenuItemEdit_UnselectAll.Click += delegate {
        foreach (TrackListViewItem item in trackListView.Items) {
          item.Checked = false;
        }
      };

      foreach (ToolStripItem item in toolStripMenuItemEdit.DropDownItems) {
        editToolStripItems.Add(item);
      }

      contextMenuStripTrackList.Items.Add(contextMenuDummyItem);

      contextMenuStripTrackList.Opening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(editToolStripItems.ToArray());
      };

      contextMenuStripTrackList.Closed += delegate {
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems.ToArray());
      };

      toolStripMenuItemEdit.DropDownOpening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems.ToArray());
      };

      toolStripMenuItemEdit.DropDownClosed += delegate {
        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(editToolStripItems.ToArray());
      };

      var initialExtractionConfig = new ExtractionConfig();

      initialExtractionConfig.Product = sortedByReleaseDateProducts[0];
      initialExtractionConfig.BgmSourcePath = sortedByReleaseDateProducts[0].DefaultBgmSourcePath;
      initialExtractionConfig.OutputDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);

      SetExtractionConfig(initialExtractionConfig, false);

      initialized = true;
    }

    public void ExtractOnShown(ExtractionConfig extractionConfig)
    {
      SetExtractionConfig(extractionConfig, true);

      doExtractOnShown = true;
    }

    protected override void OnShown(EventArgs e)
    {
      base.OnShown(e);

      if (doExtractOnShown) {
        isEdittingExtractionConfigDirty = true;
        Extract();
      }
      else {
        isEdittingExtractionConfigDirty = false;

        foreach (var track in edittingExtractionConfig.Tracks) {
          track.ClearDirty();
        }
      }
    }

    private void toolStripMenuItemFile_Quit_Click(object sender, EventArgs e)
    {
      Close();
    }

    protected override void OnClosing(CancelEventArgs e)
    {
      if (backgroundWorkerExtraction.IsBusy) {
        var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_OnClosingAbortExtraction"), "抜き出しの中断", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2);

        if (result == DialogResult.Yes) {
          // abort
          backgroundWorkerExtraction.CancelAsync();

          // wait for RunWorkerCompleted
          while (backgroundWorkerExtraction.IsBusy) {
            Application.DoEvents();
            Thread.Sleep(250);
          }
        }
        else {
          // continue extraction
          e.Cancel = true;
        }
      }

      e.Cancel |= !CheckAndConfirmSaveExtractionConfig(true);

      base.OnClosing(e);
    }

    private void SetSelectedEncoderConfigCheckedState()
    {
      var selectedEncoderConfig = edittingExtractionConfig.EncoderConfig ?? EncoderConfig.Empty;

      foreach (var item in toolStripMenuItemExtract_EncoderConfigs.DropDownItems) {
        if (item is EncoderConfigToolStripMenuItem) {
          var i = item as EncoderConfigToolStripMenuItem;

          i.Checked = (selectedEncoderConfig == i.EncoderConfig);
        }
      }
    }

    private void toolStripMenuItemTrackList_EditCollective_Click(object sender, EventArgs e)
    {
      using (var dialog = new CollectiveEffectEditDialog()) {
        var result = dialog.ShowDialog(this);

        if (result != DialogResult.OK)
          return;

        foreach (var track in trackListView.Tracks) {
          track.Effect.FadingCurveType = dialog.FadingCurveType;

          switch (dialog.EffectLengthBasis) {
            case EffectLengthBasis.FadingStart:
              track.Effect.SetLength(dialog.FadingStart, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TotalLength:
              track.Effect.SetTotalLength(dialog.TotalLength, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TimesToRepeat:
              track.Effect.SetLength(track.TrackInfo, dialog.TimesToRepeat, dialog.FadingLength, dialog.FadingOffset, dialog.TrailingSilenceLength);
              break;
          }
        }
      }
    }

#region "open/save"
    private void toolStripMenuItemFile_Open_BmgSource_Click(object sender, EventArgs e)
    {
      productSelector.ShowChangeBmgSourcePathDialog();
    }

    private void toolStripMenuItemFile_Open_ExtractionConfig_Click(object sender, EventArgs e)
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = currentConfigFilePath;
        dialog.InitialDirectory = string.IsNullOrEmpty(currentConfigFilePath) ? Environment.CurrentDirectory : Path.GetDirectoryName(currentConfigFilePath);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) == DialogResult.OK)
          LoadExtractionConfig(dialog.FileName);
      }
    }

    private void toolStripMenuItemFile_SaveExtractionConfigAs_Click(object sender, EventArgs e)
    {
      SaveExtractionConfigAs();
    }

    private void toolStripMenuItemFile_SaveExtractionConfig_Click(object sender, EventArgs e)
    {
      SaveExtractionConfig();
    }
#endregion

#region "save/load extraction config"
    private bool CheckAndConfirmSaveExtractionConfig(bool checkAllConfig)
    {
      if (!initialized)
        return true;

      var areTracksDirty = false;

      foreach (var track in edittingExtractionConfig.Tracks) {
        if (track.IsDirty) {
          areTracksDirty = true;
          break;
        }
      }

      if (!(areTracksDirty || (checkAllConfig && isEdittingExtractionConfigDirty)))
        return true;

      var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndConfirmSaveExtractionConfig"), "抜き出し設定の保存", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

      switch (result) {
        case DialogResult.Yes:
          return SaveExtractionConfig() == DialogResult.OK;

        case DialogResult.No:
          return true;

        default:
          return false; // cancel
      }
    }

    public void LoadExtractionConfig(string path)
    {
      try {
        if (!CheckAndConfirmSaveExtractionConfig(true))
          return;

        SetExtractionConfig(ExtractionConfig.LoadFrom(path), true);

        currentConfigFilePath = path;

        isEdittingExtractionConfigDirty = false;
      }
      catch (InvalidDataException ex) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_InvalidDataException", ex.InnerException.Message), "抜き出し設定ファイルの読み込みエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
        Console.Error.WriteLine(ex);
      }
      catch (NotSupportedException) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_NotSupportedException"), "抜き出し設定ファイルの読み込みエラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    }

    private void SetExtractionConfig(ExtractionConfig newExtractionConfig, bool setTracks)
    {
      var propertyChanging = new PropertyChangingEventHandler(delegate(object sender, PropertyChangingEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "BgmSourcePath") {
          lastBgmSourcePath = config.BgmSourcePath;
        }
        else if (e.PropertyName == "Product") {
          if (string.IsNullOrEmpty(productSelector.BgmSourcePath) || productSelector.BgmSourcePath == config.Product.DefaultBgmSourcePath)
            productSelector.BgmSourcePath = string.Empty;

          config.BgmSourcePath = lastBgmSourcePath;

          if (CheckAndConfirmSaveExtractionConfig(false)) {
            config.BgmSourcePath = productSelector.BgmSourcePath;

            // force save-as in next time
            currentConfigFilePath = null;

            return; // extraction config is not dirty or saved
          }
          else /* canceled */ {
            // revert bgm source and product
            productSelector.SelectedProduct = config.Product;
            productSelector.BgmSourcePath = config.BgmSourcePath;
            player.BgmSourcePath = config.BgmSourcePath;

            (e as CancellablePropertyChangingEventArgs).Cancel = true;
          }
        }
      });

      var propertyChanged = new PropertyChangedEventHandler(delegate(object sender, PropertyChangedEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "Product" && string.IsNullOrEmpty(productSelector.BgmSourcePath))
          productSelector.BgmSourcePath = config.Product.DefaultBgmSourcePath;
      });

      edittingExtractionConfig.PropertyChanging -= propertyChanging;
      edittingExtractionConfig.PropertyChanged -= propertyChanged;

      productSelector.SelectedProduct = newExtractionConfig.Product;
      productSelector.BgmSourcePath = newExtractionConfig.BgmSourcePath;

      if (setTracks) {
        trackListView.SetProductAndTracks(newExtractionConfig.Product, newExtractionConfig.Tracks);
      }
      else {
        trackListView.Product = newExtractionConfig.Product;

        newExtractionConfig.Tracks = trackListView.Tracks;
      }

      toolStripMenuItemExtract_PrependTrackNumber.Checked = newExtractionConfig.PrependTrackNumber;
      toolStripMenuItemExtract_UseWaveDashInsteadOfFullWidthTilde.Checked = newExtractionConfig.UseWaveDashInsteadOfFullWidthTilde;
      toolStripMenuItemExtract_UseTwoSpacesInsteadOfIdeographicSpace.Checked = newExtractionConfig.UseTwoSpacesInsteadOfIdeographicSpace;

      toolStripMenuItemExtract_EncoderConfigs_SaveSourceAfterEncode.Checked = newExtractionConfig.SaveSourceAfterEncode;

      edittingExtractionConfig = newExtractionConfig;

      edittingExtractionConfig.PropertyChanging += propertyChanging;
      edittingExtractionConfig.PropertyChanged += propertyChanged;

      SetSelectedEncoderConfigCheckedState();
    }

    private DialogResult SaveExtractionConfig()
    {
      if (string.IsNullOrEmpty(currentConfigFilePath)) {
        return SaveExtractionConfigAs();
      }
      else {
        edittingExtractionConfig.Save(currentConfigFilePath);

        foreach (var track in edittingExtractionConfig.Tracks) {
          track.ClearDirty();
        }

        return DialogResult.OK;
      }
    }

    private DialogResult SaveExtractionConfigAs()
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        var initialFilePath = string.IsNullOrEmpty(currentConfigFilePath)
          ? Path.Combine(Environment.CurrentDirectory, Smdn.IO.PathUtils.RemoveInvalidFileNameChars(productSelector.SelectedProduct.Title + ExtractionConfig.FileExtension))
          : currentConfigFilePath;

        dialog.FilterIndex = 0;
        dialog.FileName = initialFilePath;
        dialog.InitialDirectory = Path.GetDirectoryName(initialFilePath);
        dialog.RestoreDirectory = true;

        var result = dialog.ShowDialog(this);

        if (result == DialogResult.OK) {
          currentConfigFilePath = dialog.FileName;
          edittingExtractionConfig.Save(currentConfigFilePath);
        }

        return result;
      }
    }
#endregion

#region "drag'n drop"
    protected override void OnDragEnter(DragEventArgs drgevent)
    {
      base.OnDragEnter(drgevent);

      if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
        drgevent.Effect = DragDropEffects.All;
      else
        drgevent.Effect = DragDropEffects.None;
    }

    protected override void OnDragDrop(DragEventArgs drgevent)
    {
      try {
        var files = (string[])drgevent.Data.GetData(DataFormats.FileDrop, false);

        if (0 < files.Length) {
          if (Smdn.IO.PathUtils.AreExtensionEqual(files[0], ExtractionConfig.FileExtension))
            LoadExtractionConfig(files[0]);
          else
            productSelector.SetBgmSourcePath(files[0], true);
        }
      }
      catch {
        // ignore exceptions
      }
      finally {
        base.OnDragDrop(drgevent);
      }
    }
#endregion

#region "preparation for extracting"
    private bool CheckAndAlertBmgSourceExistence()
    {
      if (File.Exists(edittingExtractionConfig.BgmSourcePath))
        return true;

      if (string.IsNullOrEmpty(edittingExtractionConfig.BgmSourcePath))
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotSet"), "音楽ファイルが見つかりません", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
      else
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotFound", edittingExtractionConfig.BgmSourcePath), "音楽ファイルが見つかりません", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

      productSelector.Focus();
      productSelector.FocusBgmSourcePath();

      return false;
    }

    private bool CheckAndConfirmCreateOutputDirectory()
    {
      if (Directory.Exists(edittingExtractionConfig.OutputDirectory))
        return true;

      var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndConfirmCreateOutputDirectory", edittingExtractionConfig.OutputDirectory), "出力先ディレクトリの作成", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

      if (result == DialogResult.Yes) {
        Directory.CreateDirectory(edittingExtractionConfig.OutputDirectory);
        return true;
      }

      return false;
    }

    private bool CheckAndConfirmOverwriteOutputFiles(IEnumerable<Track> tracks, out Dictionary<Track, string> extractedFilePaths, out Dictionary<Track, string> encodedFilePaths)
    {
      extractedFilePaths = new Dictionary<Track, string>();
      encodedFilePaths = new Dictionary<Track, string>();

      foreach (var track in tracks) {
        var outputFileName = track.TrackInfo.GetUnifiedTitle(edittingExtractionConfig.TitleUnificationOptions);

        if (edittingExtractionConfig.PrependTrackNumber)
          outputFileName = string.Format("{0:D2}. {1}", track.TrackInfo.TrackNumber, outputFileName);

        // 無効な文字は半角空白に置き換える
        outputFileName = Smdn.IO.PathUtils.ReplaceInvalidFileNameChars(outputFileName, " ");

        extractedFilePaths.Add(track, Path.Combine(edittingExtractionConfig.OutputDirectory, outputFileName + ".wav"));

        if (!EncoderConfig.IsNullOrEmpty(edittingExtractionConfig.EncoderConfig))
          encodedFilePaths.Add(track, Path.Combine(edittingExtractionConfig.OutputDirectory, outputFileName + "." + edittingExtractionConfig.EncoderConfig.OutputExtension));
      }

      var fileExists = false;

      foreach (var outputPathLists in new[] {extractedFilePaths.Values, encodedFilePaths.Values}) {
        foreach (var outputFile in outputPathLists) {
          if (File.Exists(outputFile)) {
            fileExists = true;
            break;
          }
        }

        if (fileExists)
          break;
      }

      if (fileExists && DialogResult.No == MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndConfirmOverwriteOutputFiles"), "ファイルがすでに存在します", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2))
        return false;

      return true; // overwrite if exists
    }

    private string SelectOutputDirectory(string initialOutputDirectory)
    {
      if (string.IsNullOrEmpty(initialOutputDirectory)) {
        var dirname = Smdn.IO.PathUtils.ReplaceInvalidPathChars(edittingExtractionConfig.Product.GetUnifiedTitle(edittingExtractionConfig.TitleUnificationOptions), " ");

        initialOutputDirectory = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.Desktop), dirname);
      }

      using (var dialog = new FolderBrowserDialog()) {
        dialog.Description = ThbgmExtractor.GetText("FolderBrowserDialog_SelectOutputDirectory_Description");
        dialog.RootFolder = Environment.SpecialFolder.Desktop;
        dialog.SelectedPath = initialOutputDirectory;
        dialog.ShowNewFolderButton = true;

        if (dialog.ShowDialog(this) == DialogResult.OK)
          return dialog.SelectedPath;
        else
          return null; // as canceled
      }
    }

    private void toolStripMenuItemExtract_Extract_Click(object sender, EventArgs e)
    {
      Extract();
    }

    private void Extract()
    {
      if (!CheckAndAlertBmgSourceExistence())
        return;

      var outputDirectory = SelectOutputDirectory(edittingExtractionConfig.OutputDirectory);

      if (outputDirectory == null)
        return; // canceled

      edittingExtractionConfig.OutputDirectory = outputDirectory;

      if (!CheckAndConfirmCreateOutputDirectory())
        return;

      var extractTracks = new List<Track>();

      foreach (var track in edittingExtractionConfig.Tracks) {
        if (track.Checked)
          extractTracks.Add(track.Clone());
      }

      Dictionary<Track, string> extractedFilePaths;
      Dictionary<Track, string> encodedFilePaths;

      if (!CheckAndConfirmOverwriteOutputFiles(extractTracks, out extractedFilePaths, out encodedFilePaths))
        return;

      // begin extraction
      toolStripMenuItemExtract_Extract.Enabled = false;

      var extractionConfig = edittingExtractionConfig.Clone();

      extractionConfig.Tracks = extractTracks;

      extractionConfig.ExtractedFilePaths.Clear();

      foreach (var pair in extractedFilePaths) {
        extractionConfig.ExtractedFilePaths.Add(pair.Key, pair.Value);
      }

      extractionConfig.EncodedFilePaths.Clear();

      foreach (var pair in encodedFilePaths) {
        extractionConfig.EncodedFilePaths.Add(pair.Key, pair.Value);
      }

      progressBarExtract.Minimum = 0;
      progressBarExtract.Maximum = extractTracks.Count * 100;

      progressBarEncode.Minimum = 0;
      progressBarEncode.Maximum = extractTracks.Count;

      splitContainerExtraction.Panel2Collapsed = false;

      backgroundWorkerExtraction.RunWorkerAsync(extractionConfig);
    }
#endregion

    private void buttonAbortExtraction_Click(object sender, EventArgs e)
    {
      // abort
      if (!backgroundWorkerExtraction.IsBusy)
        return;

      backgroundWorkerExtraction.CancelAsync();
    }

#region "backgroundWorkerExtraction"
    private void backgroundWorkerExtraction_DoWork(object sender, DoWorkEventArgs e)
    {
      var config = e.Argument as ExtractionConfig;
      var context = Extractor.StartExtract(config);

      backgroundWorkerExtraction.ReportProgress(0, context);

      for (;;) {
        if (!context.IsAbortRequested && backgroundWorkerExtraction.CancellationPending) {
          e.Cancel = true;
          Extractor.Abort(context);
        }

        if (Extractor.WaitForExtractionFinished(context, 500))
          break;
        else
          backgroundWorkerExtraction.ReportProgress(0, context);
      }

      backgroundWorkerExtraction.ReportProgress(0, context);

      e.Result = context;
    }
    
    private void backgroundWorkerExtraction_ProgressChanged(object sender, ProgressChangedEventArgs e)
    {
      var context = e.UserState as ExtractionContext;
      var progress = context.Progress;

      labelExtractProgress.Text = ThbgmExtractor.GetText("labelExtractProgress_Text", progress.ExtractedTrackCount, progress.TotalTrackCount);

      if (EncoderConfig.IsNullOrEmpty(context.Config.EncoderConfig))
        labelEncodeProgress.Text = string.Empty;
      else
        labelEncodeProgress.Text = ThbgmExtractor.GetText("labelEncodeProgress_Text", progress.EncodedTrackCount, progress.TotalTrackCount);

      labelProgressStatus.Text = progress.Status;

      var perTrackProgressPercentage = (progress.TrackLength == 0L || progress.ExtractedTrackCount == progress.TotalTrackCount)
        ? 0
        : (int)(100L * progress.ExtractedLength / progress.TrackLength);

      progressBarExtract.Value = progress.ExtractedTrackCount * 100 + perTrackProgressPercentage;
      progressBarEncode.Value = progress.EncodedTrackCount;
    }

    private void backgroundWorkerExtraction_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
    {
      progressBarExtract.Maximum = 1;
      progressBarExtract.Value = 1;

      progressBarEncode.Maximum = 1;
      progressBarEncode.Value = 1;

      labelExtractProgress.Text = string.Empty;
      labelEncodeProgress.Text = string.Empty;

      if (e.Error != null) {
        Console.Error.WriteLine(e.Error);

        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_AbortedByError"), "抜き出し中断", MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
      }
      else if (e.Cancelled) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_Aborted"), "抜き出し中断", MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
      }
      else {
        var context = e.Result as ExtractionContext;
        var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_ConfirmOpenOutputDirectory"), "抜き出し完了", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

        if (result == DialogResult.Yes) {
          try {
            Smdn.IO.FileManager.Browse(context.Config.OutputDirectory);
          }
          catch {
            // ignore exceptions
          }
        }
      }

      toolStripMenuItemExtract_Extract.Enabled = true;
      splitContainerExtraction.Panel2Collapsed = true;
    }
#endregion

    private void timerLog_Tick(object sender, EventArgs e)
    {
      if (textBoxLog.TextLength == ThbgmExtractor.Log.Length)
        return;

      textBoxLog.Text = ThbgmExtractor.Log.ToString();
      textBoxLog.Select(textBoxLog.TextLength - 1, 0);
      textBoxLog.ScrollToCaret();
    }

    private bool initialized = false;
    private ExtractionConfig edittingExtractionConfig = new ExtractionConfig();
    private bool isEdittingExtractionConfigDirty = false;
    private string currentConfigFilePath = null;
    private string lastBgmSourcePath = string.Empty;
    private bool doExtractOnShown = false;
    private List<ToolStripItem> editToolStripItems = new List<ToolStripItem>();
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
  }
}